# -*- coding: utf-8 -*-
"""
Created on Wed Oct 13 09:50:47 2021

@author: lbres
"""
import pandas as pd
import geopandas as gpd
import datetime as dt
from calendar import monthrange
import numpy as np
import matplotlib.pyplot as plt

# spatiotemporal distribution of pumping
report_cell_distribution = pd.read_csv('spatial_distribution.csv', index_col=0).loc[:,'0']

# Open recharge data from the climate model
rch_zones = gpd.read_file('./climate_model/rch_zones.shp')
rch_zones['Subbasin'] = rch_zones['Subbasin'].astype(int)
# compute rch zone area
rch_zones['area'] = rch_zones['geometry'].area
rch_zones_area = rch_zones.groupby('Subbasin')['area'].sum().to_dict()

df_recharge = pd.read_csv('./climate_model/lbresin_climate_model_1981-2070_m_d-1.csv', index_col=0) #m/d
df_recharge.index = pd.to_datetime(df_recharge.index)
df_recharge.columns = df_recharge.columns.astype(int)

# compute annual total volumetric recharge per hydraulic year
df_recharge_annual = pd.DataFrame() # m/a
for year in df_recharge.index.year.unique()[:-1]:
    season_start = pd.Timestamp('{}-9-15'.format(year))
    season_end = pd.Timestamp('{}-9-14'.format(year+1))
    df_recharge_annual[year] = df_recharge[season_start:season_end].sum()
df_recharge_annual = df_recharge_annual.T
for zone in df_recharge_annual.columns: # convert to volumetric recharge 
    df_recharge_annual[zone] *= rch_zones_area[zone] # multiply with recharge area; m^3/a
df_recharge_annual = df_recharge_annual.sum(axis=1) # sum over all areas

# compute pumping scenarios (b: baseline, rnc: regional nature conservation, rri: regional rescource intensive)
df_pump_b = pd.DataFrame(columns = ['rolling_mean_total'], #m³/a
                       data = df_recharge_annual.rolling(window=5).mean()) # compute puming based on 5-year rolling mean
for report_cell in report_cell_distribution.index:
    df_pump_b.loc[:,report_cell] =  df_pump_b['rolling_mean_total'] * report_cell_distribution[report_cell] # apply spatial distribution
df_pump_b = df_pump_b.dropna()

df_pump_b.index += 1 # increase by one year, as the HSI can only make the decission to for pumping the next year


# convert to daily flux m³/d
df_pump_b_flux = pd.DataFrame(columns=report_cell_distribution.index) #m^3/d
for year in df_pump_b.index:
    df_pump_b_flux.loc[pd.Timestamp(year, 9, 15), report_cell_distribution.index] = df_pump_b.loc[year,report_cell_distribution.index] / 365.25


# rnc scenario: steady decrease to 80% till 2050
df_pump_rnc_flux = df_pump_b_flux.copy()

multiplier = pd.Series(index=df_pump_rnc_flux.index, dtype=float, data=None)
multiplier[multiplier.index < dt.datetime(2019,12,31)] = 1
multiplier[multiplier.index > dt.datetime(2050,12,31)] = 0.8
multiplier[multiplier.isna()] = np.linspace(1, 0.8, multiplier[multiplier.isna()].shape[0]+1)[1:]

df_pump_rnc_flux = df_pump_rnc_flux.apply(lambda x: x*multiplier.to_numpy(), axis=0)

# rri scenario: steady decrease to 120% till 2050
df_pump_rri_flux = df_pump_b_flux.copy()

multiplier = pd.Series(index=df_pump_rri_flux.index, dtype=float, data=None)
multiplier[multiplier.index < dt.datetime(2019,12,31)] = 1
multiplier[multiplier.index > dt.datetime(2050,12,31)] = 1.2
multiplier[multiplier.isna()] = np.linspace(1, 1.2, multiplier[multiplier.isna()].shape[0]+1)[1:]

df_pump_rri_flux = df_pump_rri_flux.apply(lambda x: x*multiplier.to_numpy(), axis=0)

# Plot result
fig, ax = plt.subplots()                     
ax.plot(df_pump_rnc_flux.sum(axis=1).groupby(df_pump_rnc_flux.index.year).sum())
ax.plot(df_pump_b_flux.sum(axis=1).groupby(df_pump_b_flux.index.year).sum())    
ax.plot(df_pump_rri_flux.sum(axis=1).groupby(df_pump_rri_flux.index.year).sum())
    
#Export result
df_pump_rri_flux.to_csv('./monthly_pumping_rri.csv') #m^3/d
df_pump_rnc_flux.to_csv('./monthly_pumping_rnc.csv') #m^3/d
df_pump_b_flux.to_csv('./monthly_pumping_b.csv') #m^3/d


