#!/home/lbresin/anaconda3/bin/python3.8 -i
# -*- coding: utf-8 -*-
"""
Created on Fri May 18 13:02:38 2018

@author: lbresin
"""

import pandas as pd
import os
import shutil
import numpy as np
import geopandas as gpd
import matplotlib.pyplot as plt

# ## Change current working directory for testing
# if input('Changing working directory (y:yes, n: no): '):
#     os.chdir("/home/lbresin/ownCloud/MedWater/hydrogeosphere/westernmountainaquifer_33_varsat_dcpe_test_zoned_rch")

# Some function to create tables for unsaturated flow
def mualem_vanGenuchten_ps_table(pressure, alpha, beta, S_wr):
    '''
    This Function calculates the water saturation (S_w) as a function of pressure (Psi), alpha, beta and the residual water saturation (S_wr)
    
    Input:
        pressure:  pressure
        alpha: The alpha value of Van Genuchten parametrization
        beta:  The beta value of Van Genuchten parametrization
        S_wr:  Residual water saturation
    '''
    nu=1-1/beta
    S_w=S_wr+(1-S_wr)*(1+abs(alpha*pressure)**beta)**(-nu)
    return S_w

def mualem_vanGenuchten_sk_table(S_w, alpha, beta, S_wr, min_rel_perm,l_p=0.5):
    '''
    This Function calculates the relative hydraulic conductivty (k_r) as a function of water saturation (S_w), alpha, beta and the residual water saturation
    
    Input:
        S_w:  Water saturation
        alpha: The alpha value of Van Genuchten parametrization
        beta:  The beta value of Van Genuchten parametrization
        S_wr:  Residual water saturation
        l_p:   pore_connectivity (i.e., 0.5 based on Mualem [1976])
        
    '''
    nu=1-1/beta #nu value as defined by VG
    S_e = (S_w - S_wr)/(1-S_wr)
    k_r=S_e**l_p*(1-(1-S_e**(1/nu))**nu)**2
    k_r[np.where(k_r < min_rel_perm)]=min_rel_perm
    return k_r       

def brooks_corey_ps_table(pressure, alpha,beta, S_wr):
    '''
    This Function calculates the water saturation (S_w) as a function of pressure (Psi), alpha, beta and the residual water saturation (S_wr)
    
    Input:
        pressure:  pressure
        alpha: The alpha value of Van Genuchten parametrization
        beta:  The beta value of Van Genuchten parametrization
        S_wr:  Residual water saturation
    '''
    if pressure < -1/alpha:
        S_w=S_wr+(1-S_wr)*abs(alpha*pressure)**(-beta)
    elif pressure >= -1/alpha:
        S_w=1
    return S_w 

def brooks_corey_sk_table(S_w, beta, S_wr, l_p=2.):
    '''
    This Function calculates the relative hydraulic conductivty (k_r) as a function of water saturation (S_w), beta and the residual water saturation
    
    Input:
        S_w:  Water saturation
        beta:  The beta value of Van Genuchten parametrization
        S_wr:  Residual water saturation
        l_p:   pore_connectivity (i.e., 2.0)
        
    '''
    S_e = (S_w - S_wr)/(1-S_wr)
    k_r=S_e**(2/beta+l_p+2)
    return k_r

# Define material properties of all zones (*.mprops file)
def write_mprops_csv(data):
    '''
    Store the current mprops parameters (i.e., data)  in the file 'mprops.csv' under the directory 'hgs_files/'
    '''
    with open(cwd+'/mprops.csv','w') as file:
        file.write(data.to_csv())
        
def write_mprops_csv_to(data, file):
    '''
    Store the current mprops parameters (i.e., data) under a defined file name'
    '''
    with open(file,'w') as file:
        file.write(data.to_csv())

def update_material_loading_file(data):
    '''
    This reads the mprops DataFrame and writes a new HGS material properties loading file. 
    This is necessary when new material zones were added or old ones were removed
    '''
    with open(cwd+'/mprops.in', 'w') as file:
        for index in data.index:
            file.write('!--------------------------zone'+str(index)+'\n')
            file.write('choose zone number\n'+str(index)+'\n')
            file.write('read properties\n')
            file.write('material_zone_'+str(index)+'\n')
            file.write('clear chosen zones\n\n')
  
def write_mprops(data, table_type='hgs', minimum_permeability=False, table_smoothness_factor=True, maximum_sk_slope=True, vg_table_entries=150):
    '''
    This function creates a mprops file for HGS based on a mprops.csv file, which contains material properties of each individual zone

    Input:
        data -> Pandas database of the material properties
        table_type -> Option when hgs_table = False; Define the table type (i.e., linear, vg, HGS_vG_min, HGS_vG_nomin)
           'linear':  		Linear retention curve based on a 'pressure-saturation-relative conductivity' relationship based on van Genuchten at P = 0 and P = -3000. Intermediate values are linear.
           'vg':		Create a table of the van-Genuchten relationship.
           'HGS_vG_min':	Create van-Genuchten table created via the already existing function of HydroGeoSphere with a minimum relative conductivity in place
           'HGS-vG_nomin':	Create van-Genuchten table created via the already existing function of HydroGeoSphere with no minimum relative conductivity in place
    '''
    shutil.move(cwd+'/mprops.csv',cwd+'/mprops.csv.prev')
    write_mprops_csv(data)
    with open(cwd+'/pm.mprops', 'w') as file: # Anisotropic original data
        for i in data.index:
            file.write('!------------------------------------------\n')
            file.write('material_zone_'+str(i))
            file.write('\n\nk isotropic\n')
            file.write('  '+'{0:.5e}'.format(data['hk'][i]).replace('e','d')+'\n\n')
            file.write('Specific storage\n')
            file.write('  '+'{0:.5e}'.format(data['Ss'][i]).replace('e','d')+'\n\n')
            file.write('Porosity\n')
            file.write('  '+'{0:.5e}'.format(data['porosity'][i]).replace('e','d')+'\n\n')
            
            ### HGS Table ###
            if table_type == 'hgs': # create table by HGS
                file.write('unsaturated van genuchten functions\n')
                file.write('  alpha\n')
                file.write('  '+'{0:.5e}'.format(data['alpha'][i]).replace('e','d')+'\n')
                file.write('  beta\n')
                file.write('  '+'{0:.5e}'.format(data['beta'][i]).replace('e','d')+'\n')
                file.write('  residual saturation\n')
                file.write('  '+'{0:.5e}'.format(data['sat_res'][i]).replace('e','d')+'\n')            
              
                # minumum relative permeability
                if minimum_permeability==True:
                    file.write('  minimum relative permeability\n')
                    file.write('  '+'{0:.5e}'.format(data['min_rel_perm'][i]).replace('e','d')+'\n')
                    
                # table smoothness factor
                if table_smoothness_factor==True:
                    file.write('  '+'table smoothness factor\n'+'  '+'{0:.5e}'.format(data['tbl_smoothness_factor'][i]).replace('e','d')+'\n')
              
                # table minimum pressure
                file.write('  '+'table minimum pressure\n'+'  '+'{0:.5e}'.format(data['tbl_min_pressure'][i]).replace('e','d')+'\n')
                
                # table maximum s-k slope
                if maximum_sk_slope==True:
                    file.write('  '+'table maximum s-k slope\n'+'  '+'{0:.5e}'.format(data['tbl_max_sk_slope'][i]).replace('e','d')+'\n')
              
                file.write('  generate tables from unsaturated functions\n')
                file.write('end ! functions\n\n')
            
            ### VG-Table ###
            elif table_type=='vg-table': # create table via Mualem and VG
                file.write('!Table created via van Genuchten relationship of saturation/pressure/relative conductivity\n')
                file.write('! Table type:\t\t{}\n'.format(table_type))
                file.write('! Alpha:\t\t'+format(data['alpha'][i]).replace('e','d')+'\n')
                file.write('! Beta:\t\t'+format(data['beta'][i]).replace('e','d')+'\n')
                file.write('! Residual Saturation:\t'+format(data['sat_res'][i]).replace('e','d')+'\n')
                file.write('! Minimum Pressure:\t{}\n'.format(data['tbl_min_pressure'][i]))
                if  minimum_permeability==True:
                    file.write('! Minimum Permeability:\t{0:.9e}\n\n'.format(data['min_rel_perm'][i]))
                else:
                    file.write('! Minimum Permeability:\t{}\n\n'.format(False))
                
                file.write('unsaturated tables\n')
                file.write('  saturation-relative k\n')
                
                # compute table
                pressure = np.append(np.geomspace(data['tbl_min_pressure'][i], -1e-2, num=vg_table_entries-1), [0], axis=0)
                #pressure=np.linspace(data['tbl_min_pressure'][i], 0, vg_table_entries)
                
                S_w = mualem_vanGenuchten_ps_table(pressure, data['alpha'][i], data['beta'][i], data['sat_res'][i])
                if  minimum_permeability==True:
                    k_r = mualem_vanGenuchten_sk_table(S_w, data['alpha'][i],data['beta'][i],data['sat_res'][i], data['min_rel_perm'][i])
                else:
                    k_r = mualem_vanGenuchten_sk_table(S_w, data['alpha'][i],data['beta'][i],data['sat_res'][i], 1e-20)
                
                for S_wi, k_ri in zip(S_w, k_r):
                    file.write('    '+'{0:.9e}'.format(S_wi)+' '+'{0:.9e}'.format(k_ri)+'\n')
                file.write('  end !sat-krw\n')
                file.write('  pressure-saturation\n')
                for pressurei, S_wi in zip(pressure, S_w):
                    file.write('    '+'{0:.9e}'.format(pressurei)+' '+'{0:.9e}'.format(S_wi)+'\n')
                file.write('  end ! p-sat\n')
                file.write('end ! unsat tables\n')
              
            ### Linear Table ###
            elif table_type=='vg-table' or table_type=='linear-table' : # create table via Mualem and VG
                file.write('!Linear Table\n')
                file.write('! Table type:\t'+table_type+'\n')
                file.write('! Alpha:\t'+format(data['alpha'][i]).replace('e','d')+'\n')
                file.write('! Beta:\t'+format(data['beta'][i]).replace('e','d')+'\n')
                file.write('! Residual Saturation:\t'+format(data['sat_res'][i]).replace('e','d')+'\n')
                file.write('unsaturated tables\n')
                file.write('  saturation-relative k\n')
     
                S_wa, S_wb=mualem_vanGenuchten_ps_table(np.array([data['tbl_min_pressure'][i], 0]),data['alpha'][i],data['beta'][i], data['sat_res'][i])
                k_ra, k_rb=mualem_vanGenuchten_sk_table(np.array([S_wa,S_wb]), data['alpha'][i],data['beta'][i],data['sat_res'][i], 0)
                pressure=np.linspace(data['tbl_min_pressure'][i],0,10)
                S_w=np.linspace(S_wa, S_wb,10)
                k_r=np.linspace(k_ra, k_rb,10)
              
                for S_wi,k_ri in zip(S_w,k_r):
                    file.write('    '+'{0:.5e}'.format(S_wi)+' '+'{0:.5e}'.format(k_ri)+'\n')
                file.write('  end !sat-krw\n')
                file.write('  pressure-saturation\n')
                for pressurei,S_wi in zip(pressure, S_w):
                    file.write('    '+'{0:.5e}'.format(pressurei)+' '+'{0:.5e}'.format(S_wi)+'\n')
                file.write('  end ! p-sat\n')
                file.write('end ! unsat tables\n')
            file.write('end material\n')
            
def undo():
    data=pd.read_csv(cwd+'/mprops.csv.prev', index_col=0)
    write_mprops(data)
    return data

def increase_min_perm():
    data.hk*=2**(data.hk.min()/data.hk)/20+0.95
    data.vk*=2**(data.vk.min()/data.vk)/20+0.95
    
def plot_hydraulic_conductivity(parameter, layer):
    '''
    This function plots the spatial extends of the individual zones. A hydraulic paramter, i.e., hk, vk, Ss,..; is visualized via a colorbar.
    
    Input:
        data - The Pandas dataframe with all hydraulic parameters
        cwd - current working directory
        parameter - Parameter that shall be ploted
        layer - geological layer that shall be ploted, i.e. Layer 4 = Overlaying Fm., Layer 3 =  Upper Aquifer, Layer 2 = Moza Fm., Layer 1 = Lower Aquifer
    '''
    plt.ion()
    model = gpd.read_file(cwd+"/zones_shp/zones_l"+str(layer)+".shp")
    model['plot_value']=[data[parameter][int(i)] for i in model.zone]
    
    ax = model.plot(column='plot_value',cmap='OrRd', edgecolor='black', legend=True, figsize=(5*1.5,7.5*1.5))
    #ax.get_legend().set_bbox_to_anchor((1.5, 1))
    
    plt.axis('equal')

    ax.set_ylabel('Y (m)')
    ax.set_xlabel('X (m)')
    ax.set_title('Layer: '+str(layer)+', Parameter: '+parameter)

    model['coords'] = model['geometry'].apply(lambda x: x.representative_point().coords[:])
    model['coords'] = [coords[0] for coords in model['coords']]
    
    for idx, row in model.iterrows():
        plt.annotate(s=row['zone'], xy=row['coords'], horizontalalignment='center')
   
    plt.show()
    
# determine working directory in the case the script was launched in the overlying HydroGeoSphere simulation folder
cwd=os.getcwd()
if 'hgs_files' not in cwd:
    cwd += '/hgs_files'
if 'material_zones' not in cwd:
    cwd += '/material_zones'

data=pd.read_csv(cwd+'/mprops.csv', index_col=0)


# Just for fun
print("\n")
print(r"""  Create porous media parameter file for HGS

                                   ._ o o
                                   \_`-)|_
                                ,""       \ 
                              ,"  ## |   ಠ ಠ. 
                            ," ##   ,-\__    `.
                          ,"       /     `--._;)
                        ,"     ## /
                      ,"   ##    /


                """)
print("\n You have the following options:\n")
print("1)\twrite_mprops_csv_to(data, file)\n")
print("\t write the mprops data to the mprops.csv file\n\n")
print("2)\twrite_mprops(data, table_type=['hgs', 'vg-table', 'linear-table'],\n\tminimum_permeability=False, table_smoothness_factor=True, maximum_sk_slope=True, vg_table_entries=250)\n")
print("\tThis function creates a mprops file for HGS based on a mprops.csv file, which contains material properties of each individual zone\n\n\tInput:\n\tdata -> Pandas database of the material properties\n\thgs_table -> generate the table with hgs table generator (True/False)\n\tmin_rel_perm -> Calculate the Van Genuchten table with a minimum relative permeability. This only works with the HGS intrinsic table generator (True/False)\n\ttable_type -> Option when hgs_table = False; Define the table type (i.e., linear, vg)\n")
print("\ti.e., write_mprops(data, table_type='vg-table', minimum_permeability=False, table_smoothness_factor=False, maximum_sk_slope=False, vg_table_entries=250)\n")
print("3)\tundo()\n")
print("\t restore previous parameters (before command 2 was run!)\n\n")
print("4)\texit()\n")
print("\t exit the interactive python env\n\n")
print("5)\tupdate_material_loading_file(data)\n")
print('\t This reads the mprops DataFrame and writes a new HGS material properties loading file.\n\t This is necessary when new material zones were added or old ones were removed\n\n')
print("6)\tplot_hydraulic_conductivity(parameter, layer)\n")
print("\t This function plots the spatial extends of the individual zones. A hydraulic paramter, i.e., hk, vk, Ss,..; is visualized via a colorbar.")



print("----------------------------------------------")

