#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May  7 12:55:46 2021

@author: lbresin
"""


import pandas as pd
from scipy.io import FortranFile
from scipy.spatial import ConvexHull
from pandarallel import pandarallel


def read_coordinates(coordinates_file):
    ''' This reads the binary HGS coordinates file (i.e.; {prefix}o.coordinates_pm or {prefix}o.coordinates_dual) and returns an m x n numpy array of the XYZ coordinates of all nodes.
    Input:
        - coordinates_file:      Path to the coordinate file
    Output:
        - nn_xyz:                m x n Numpy array of the node coordinates (m = number of nodes, n = three dimensions, i.e., x y z) '''
    fid = FortranFile(coordinates_file)
    n = fid.read_ints()[0]    
    
    df = pd.DataFrame(data = fid.read_reals().reshape(n,3))
    df.columns = ['x', 'y', 'z']
    df.index += 1
    
    return df

def read_heads(head_file):
    ''' This reads the binary HGS head file (i.e.; {prefix}o.head_pm.**** or {prefix}o.head_dual.****) and returns an m x 1 numpy array of the simulated heads of all nodes.
    Input:
        - coordinates_file:      Path to the head file
    Output:
        - nn_xyz:                m x 1 Numpy array of the simulates heads (m = number of nodes, n = 1) '''
    fid = FortranFile(head_file)
    _ = fid.read_ints()
    return fid.read_reals()

def read_saturation(sat_file):
    ''' This reads the binary HGS head file (i.e.; {prefix}o.sat_pm.**** or {prefix}o.sat_dual.****) and returns an m x 1 numpy array of the simulated heads of all elements.
    Input:
        - coordinates_file:      Path to the head file
    Output:
        - nn_xyz:                m x 1 Numpy array of the simulates heads (m = number of nodes, n = 1) '''
    fid = FortranFile(sat_file)
    _ = fid.read_ints()#dtype='int32')
    return fid.read_reals(dtype='float32')  

def read_elements(element_file):
    ''' This reads the binary HGS head file (i.e.; {prefix}o.sat_pm.**** or {prefix}o.sat_dual.****) and returns an m x 1 numpy array of the simulated heads of all elements.
    Input:
        - coordinates_file:      Path to the head file
    Output:
        - nn_xyz:                m x 1 Numpy array of the simulates heads (m = number of nodes, n = 1) '''
    fid = FortranFile(element_file)
    a = fid.read_ints(dtype='int32')[0]
    b = fid.read_ints(dtype='int32')[0]
    
    c = fid.read_ints(dtype='int32')
    c = c.reshape([b,a])
    
    d = pd.DataFrame(index=range(1,b+1), columns=['node_list'])
    d.loc[:, 'node_list'] = list(c)
    return d

def get_output_times_from_eco(file_dir):
    with open(file_dir, 'r') as fid:
        lines = fid.readlines()

        for i in reversed(range(len(lines))):
            if 'EXIT OUTPUT TIME' in lines[i]:
                end = i - 1 
        for i in reversed(range(end)):
            if 'output times' in lines[i]:
                start = i + 1 
        output_times = pd.DataFrame(columns=['time'], data=map(lambda x: x.split()[-1],lines[start:end]))
        output_times.index = output_times.index.astype(int) + 1 
    
        return output_times.astype(float)

if __name__ == '__main__':
    ### SIMULATION PARAMETER & SETTING ###
    prefix = 'transient'

    # Open Coordinates
    df_nodes =  read_coordinates('{prefix:}o.coordinates_pm'.format(prefix=prefix))
    df_nodes['node_number'] = df_nodes.index
    idx_nodes_top = df_nodes.groupby(['x','y'])['z'].transform(max) == df_nodes['z']
    idx_nodes_bottom = df_nodes.groupby(['x','y'])['z'].transform(min) == df_nodes['z']
                      
    # Open Elements
    df_elements = read_elements('transiento.elements_pm')
    tmp = pd.read_csv('element_zones', delim_whitespace=True, header=None)
    tmp.columns = ['element', 'zone']
    tmp = tmp.set_index('element')
    df_elements = df_elements.merge(tmp, left_index=True, right_index=True, how='outer')
    del tmp
                               
    #Compute elemental volume
    pandarallel.initialize()
    def compute_volume(x):
        vol = ConvexHull(list(zip(df_nodes.loc[x, 'x'],
                                      df_no des.loc[x, 'y'],
                                      df_nodes.loc[x, 'z'])))
        return vol.volume
    df_elements['volume'] = df_elements.node_list.parallel_apply(compute_volume)
    
    # Find max_z
    def find_max_z(x):
        return df_nodes.loc[x, 'z'].max()
    df_elements['max_z'] = df_elements.node_list.parallel_apply(find_max_z)
    def find_min_z(x):
        return df_nodes.loc[x, 'z'].min()
    df_elements['min_z'] = df_elements.node_list.parallel_apply(find_min_z)
    
    # Figure out simsteps
    out_times = get_output_times_from_eco('transiento.eco')
    out_times['date'] = out_times['time'].apply(lambda x: pd.Timestamp(1950,9,15) + pd.Timedelta(x, unit='days'))
    
    # Compute vadose water volume in coastal aquifer   
    df_vadose_water_coastal_aq = pd.DataFrame(index=out_times['date'], columns=['vadose_store'])
    for sim_step in out_times.index:
        df_nodes['saturation'] = read_saturation('{prefix:}o.sat_pm.{sim_step:04d}'.format(prefix=prefix, sim_step=sim_step))
        
        def sat_lookup(x):
            return df_nodes.loc[x, 'saturation'].median()
        df_elements['saturation'] = df_elements.node_list.parallel_apply(sat_lookup)
        
        df_vadose_water_coastal_aq.loc[out_times.loc[sim_step, 'date'], 
                                       'vadose_store'] = df_elements.loc[((df_elements.saturation<0.995) & (df_elements.zone >= 71)), ['saturation', 'volume']].product(axis=1).sum() * 0.05      

    df_vadose_water_coastal_aq.to_csv('vadose_water_in_coastal_aq.csv')
    
    print('Water stored below 4m: {0:1.15e}'.format(df_elements.loc[df_elements['max_z']<=4, 'volume'].sum()*0.05))
    print('Vadose water in the coastal: {0:1.15e}'.format(df_elements.loc[((df_elements.saturation<0.995) & (df_elements.zone >= 71)), ['saturation', 'volume']].product(axis=1).sum() * 0.05))
